
/*!
 * bundle
 * Copyright(c) 2011 LearnBoost <dev@learnboost.com>
 * MIT Licensed
 */

/**
 * Library version.
 */
exports.version = '0.0.1';

/**
 * Module dependencies.
 */
var jade = require('jade')
  , fs = require('fs')
  , exec = require('child_process').exec
  , Log = require('log')


 /**
 * Exports.
 */
exports.Bundle = Bundle;

exports.createBundle = function (opts) {
  return new Bundle(opts);
};


/**
 * Bundle constructor
 *
 * options parameters:
 * - views: templates files default folder
 * - tmp: temporal proccessing folder
 * - gzip: gzip command
 * - engine: view engine
 * - extension: extension file
 * - encoding
 * - tmpPackname: temporal package name
 * - addLogFile: defines add or not a log file
 *
 * @api private
 */
function Bundle (opts) {
  // default options
  this.options = {
      views: 'views'
    , layouts: 'views/layouts/'
    , tmp: '/tmp'
    , gzip: 'zip'
    , engine: 'jade'
    , extension: 'html'
    , encoding: 'utf8'
    , tmpPackname: 'bundle'
    , addLogFile: true
    , depth: 0
  };

  // mixing options
  for(var k in opts)
    this.options[k] = opts[k];
  
  // initial configs
  this.generated = false;
  this.file = null;
  this.files = {};
  this.filesCount = 0;

  this.hash = 'node-bundle-' + String(Math.random()).substr(2);
};


/**
 * Define a file
 *
 * @api private
 */
Bundle.prototype.addFile = function (name, callback) {
  var args = arguments
    , name = args[0]
    , fn = typeof args[1] == 'function' ? args[1] : args[2]
    , options = typeof[1] == 'object' ? args[1] : {}

  this.files[name] = {fn:fn, options: options};
  this.filesCount++;
};


/**
 * Generates the structure in a tmp location and gzips it
 *
 * @api private
 */
Bundle.prototype.generate = function (params, callback) {
  var self = this;

  // options alias
  var _ops = this.options;

  var createTreePackage = function(fn) {
    // create tmpbundle folder
    self.bundlePath = _ops.tmp + '/' + self.hash;
    self.createFolder(self.bundlePath, function() {
      // log
      self.log('create tmp folder: ' + self.bundlePath);

      fn();
    });

  }

  var package = function (callback) {
    // sets this.generated = true and this.file = to location
    self.generated = true;

    // make compressed package
    self.file = self.bundlePath + '/' + _ops.tmpPackname + '.zip';
    var zipCmd = '(cd ' + self.bundlePath + ' ; '
                + 'zip -r ' + _ops.tmpPackname + '.zip ./ )'

    self.log('compresing: ' + zipCmd);

    exec(zipCmd, function (err, stdout, stderr) {
      if (err) return callback(err);
      callback(null, self.file);
    });
  }

  var renderLayout = function(layout, opts, html, callback) {
    if(!layout) return callback(null);
    delete opts.layout;
    
    var layoutFilename =_ops.layouts + layout
      , options = {
          body: html
        }

    // mixing options
    for(var k in opts) options[k] = opts[k];
    
    jade.renderFile(layoutFilename, { locals: options }, function(err, html) {
      if(err) return callback(err);

      self.log('rendering layout: ' + layoutFilename);
      return callback(null, html);
    });
  }

  // if this.generated = true calls immediately with this.file
  // gzips it
  // calls the function with the file location
  if(self.generated){
    callback(null, self.file);
  }
  else {
    createTreePackage(function() {
      var fileCounter = self.filesCount;

      // outputs everything to a temporary directory (by calling all the file
      // routes in paralell and keeping a counter of completion)
      for (var filename in self.files) {

        (function (file) {
          
          // execute file function
          file.fn(params, function(template, options) {
            var viewsFilename =_ops.views + template;

            // render html file
            jade.renderFile(viewsFilename, { locals: options }, function(err, html) {
              if (err) return callback(err);
              self.log('render file: ' + viewsFilename);

              renderLayout(file.options.layout, options, html, function(err, htmlLayout) {
                if (err) return callback(err);

                // replace html by htmlLayout
                if(htmlLayout) {
                  html = htmlLayout;
                  self.log('decoration: insert html into layout');
                }
              
                // create folders for file ?
                var rePath = file.name.match(/(.+)\/(\w*$)/)
                  , addFolder = rePath ? self.bundlePath + '/' + rePath[1] : false;

                self.createFolder(addFolder, function(folder) {

                var saveFilename = self.bundlePath + '/' + file.name + '.' + _ops.extension;

                  // save html in file
                  fs.writeFile(saveFilename, html, _ops.encoding, function (err) {

                    if (err) return callback(err);

                    self.log('save file: ' + saveFilename);

                    --fileCounter || package(callback);
                  });

                });
              });
            });
          });

        })({name: filename, fn: self.files[filename].fn, options: self.files[filename].options });

      };

    });

  }
};


/**
 * Generates a bundle and saves it to a file
 *
 * @api private
 */
Bundle.prototype.save = function (params, file, callback) {
  var self = this;
  this.generate(params, function(){

    var withPath = file.match(/[\w|\/].*\//)
      , path = withPath.length ? withPath[0].substr(0, withPath[0].length - 1) : null;

    self.createFolder(path, function() {

      var mvFileCmd = 'mv ' + self.file + ' ' + file;
      exec(mvFileCmd, function (err, stdout, stderr) {
        if(err) return callback(err);

        self.log('moving file: ' + mvFileCmd);

        callback(null, file);
      });

    });

  });
};


/**
 * Generates a bundle and gets the buffer
 *
 * @param private
 */
Bundle.prototype.get = function (params, callback) {
  var self = this;
  this.generate(params, function() {

    // reads the file ...
    fs.readFile(self.file, function(err, data) {
      if(err) return callback(err);

      // sets this.generated = false
      self.generated = false;

      // removes the file
      var rmTmpCmd = 'rm ' + self.bundlePath + ' -R';
      exec(rmTmpCmd, function (err, stdout, stderr) {
        if(err) return callback(err);
        
        // ... and passes the buffer to fn
        callback(null, data);
      });
      
    });

  });

};


/**
 * Log util method
 **/
Bundle.prototype.log = function (msg) {
  if(!this.options.addLogFile) return null;
  if(!this.objLog)
    this.objLog = new Log(Log.INFO, fs.createWriteStream(this.bundlePath + '/' + this.options.tmpPackname + '.log'));
  
  this.objLog.info(msg);
}


/**
 * createFolder() util method
 **/
Bundle.prototype.createFolder = function(path, fn) {
  if(!path) return fn('');

  var mkdirCmd = 'mkdir ' + path + ' -p';

  exec(mkdirCmd, function (err, stdout, stderr) {
    if(err) return fn(err);
    fn(path);
  });

}


Bundle.prototype.cleanTmpFolder = function (callback) {

  var rmTmpCmd = 'rm ' + this.bundlePath + ' -R';
  exec(rmTmpCmd, function(err, stdout, stderr) {
    if(err) return callback(err);
    if(stderr) return callback(err, stderr);

    if(callback) callback();
  })

}


/**
 * Escape the given shell `arg`.
 *
 * @param {String} arg
 * @return {String}
 * @api public
 */
exports.escape = function (arg) {
  return '"' + String(arg).replace(/"/g, '\\"') + '"';
};
